:- use_module(library(clpfd)).
:- use_module('/afs/inria.fr/rocq/home/contraintes/jmartin/interpreter/prolog/r2cp').
:- init_interpreter.


q(I, _A_) :- 
	rcp_var(from(q(I),0,1), Q_I),
	rcp_rec([row=Q_I,column=I], _A_).

q_search(I, _A_) :- 
	rcp_var(from(q(I),0,1), Q_I),
	rcp_rec([row=Q_I,column=I], _A_).

board_foldr_1([], _, []) :- 
	!.
board_foldr_1([I_2|Tail], [N], [_A_|_B_]) :- 
	board_foldr_1(Tail, [N], _B_),
	q(I_2, _A_).

board_foldr_1_search([], _, []) :- 
	!.
board_foldr_1_search([I_2|Tail], [N], [_A_|_B_]) :- 
	board_foldr_1_search(Tail, [N], _B_),
	q_search(I_2, _A_).

board(N, _B_) :- 
	rcp_range([1..N], _A_),
	board_foldr_1(_A_, [N], _B_).

board_search(N, _B_) :- 
	rcp_range([1..N], _A_),
	board_foldr_1_search(_A_, [N], _B_).

safe_foldl_2([], I_and, _, I_and) :- 
	!.
safe_foldl_2([R_7|Tail], I_and, [L,Q_3], _P_) :- 
	rcp_att(column, Q_3, _A_),
	I_10=_A_,
	rcp_att(column, R_7, _B_),
	J_12=_B_,
	fd_lt(I_10, J_12, _C_),
	rcp_att(row, Q_3, _D_),
	fd_aeq(J_12-I_10, _E_),
	rcp_att(row, R_7, _F_),
	fd_aeq(_E_+_F_, _G_),
	fd_neq(_D_, _G_, _H_),
	rcp_att(row, Q_3, _I_),
	fd_aeq(I_10-J_12, _J_),
	rcp_att(row, R_7, _K_),
	fd_aeq(_J_+_K_, _L_),
	fd_neq(_I_, _L_, _M_),
	fd_and(_H_, _M_, _N_),
	fd_implies(_C_, _N_, _O_),
	fd_and(I_and, _O_, _Q_),
	safe_foldl_2(Tail, _Q_, [L,Q_3], _P_).

safe_foldl_2_search([], I_and, _, I_and) :- 
	!.
safe_foldl_2_search([R_7|Tail], I_and, [L,Q_3], _L_) :- 
	rcp_att(column, Q_3, _A_),
	I_10=_A_,
	rcp_att(column, R_7, _B_),
	J_12=_B_,
	(
	fd_lt(I_10, J_12),
	rcp_att(row, Q_3, _C_),
	fd_aeq(J_12-I_10, _D_),
	rcp_att(row, R_7, _E_),
	fd_aeq(_D_+_E_, _F_),
	rcp_att(row, Q_3, _G_),
	fd_aeq(I_10-J_12, _H_),
	rcp_att(row, R_7, _I_),
	fd_aeq(_H_+_I_, _J_),
	fd_equiv((fd_neq(_C_,_F_),fd_neq(_G_,_J_)), _K_, 1)
	;
	fd_geq(I_10, J_12),
	fd_aeq(_K_, 1)
	),
	rcp_fol_call(I_and),
	rcp_fol_call(_K_),
	safe_foldl_2_search(Tail, _M_, [L,Q_3], _L_).

safe_foldl_1([], I_and, _, I_and) :- 
	!.
safe_foldl_1([Q_3|Tail], I_and, [L], _B_) :- 
	safe_foldl_2(L, 1, [L,Q_3], _A_),
	fd_and(I_and, _A_, _C_),
	safe_foldl_1(Tail, _C_, [L], _B_).

safe_foldl_1_search([], I_and, _, I_and) :- 
	!.
safe_foldl_1_search([Q_3|Tail], I_and, [L], _B_) :- 
	safe_foldl_2_search(L, 1, [L,Q_3], _A_),
	rcp_fol_call(I_and),
	rcp_fol_call(_A_),
	safe_foldl_1_search(Tail, _C_, [L], _B_).

safe(L, _B_) :- 
	variables(L, _A_),
	all_different(_A_),
	safe_foldl_1(L, 1, [L], _B_).

safe_search(L, _C_) :- 
	variables(L, _A_),
	all_different(_A_),
	safe_foldl_1_search(L, 1, [L], _B_),
	rcp_fol_call(_B_),
	fd_aeq(_C_, 1).

queens_constraints(N, _C_) :- 
	board(N, _A_),
	B_13=_A_,
	variables(B_13, _B_),
	domain(_B_, 1, N),
	safe(B_13, _C_).

queens_constraints_search(N, _D_) :- 
	board_search(N, _A_),
	B_13=_A_,
	variables(B_13, _B_),
	domain(_B_, 1, N),
	safe_search(B_13, _C_),
	rcp_fol_call(_C_),
	fd_aeq(_D_, 1).

var_order_criterion(1, Candidate, [], _C_) :- 
	rcp_call(Candidate, _A_),
	rcp_att(row, _A_, _B_),
	fd_size(_B_, _C_).

queens_search(N, _C_) :- 
	board(N, _A_),
	B_14=_A_,
	variables(B_14, _B_),
	variable_ordering([least(var_order_criterion(1,[]))]),
	rcp_labeling(_B_),
	fd_aeq(_C_, 1).

queens_search_search(N, _C_) :- 
	board_search(N, _A_),
	B_14=_A_,
	variables(B_14, _B_),
	variable_ordering([least(var_order_criterion(1,[]))]),
	rcp_labeling(_B_),
	fd_aeq(_C_, 1).

solve([N]) :- 
	rcp_fol_call(_C_),
	N=4,
	queens_constraints(N, _A_),
	queens_search(N, _B_),
	fd_and(_A_, _B_, _C_).
